package netgame.draw;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.URL;
import javax.imageio.ImageIO;
import javax.swing.*;
import netgame.common.*;

/**
 * This subclass of JPanel defines a client for the NetDraw
 * application.  (To have a complete application, run NewDrawMain,
 * which opens a window that shows a DrawPanel.)
 * The NetDraw application allows uses to work together to
 * create simple drawings.  Everything drawn by one user
 * is transmitted to the NetDraw server (or "hub"), which
 * then forwards a copy to all connected clients (including
 * the client that sent that item to the server).  Every 
 * client who gets the message -- except the one who sent it --
 * can then add the item to its own picture.  This means
 * that an item drawn by one client is added to every 
 * client's picture.  Items are transmitted as strings that 
 * describe the thing that was drawn.  Only two types of item 
 * are supported: Lines and stamps.
 *    A line has a color and two endpoints (x1,y1) and (x2,y2).
 * The string for a line consists of the word "line" followed
 * by seven numbers, all separated by spaces.  The seven numbers
 * are the R, G, and B values of the color and the coordinates
 * x1, y1, x2, adn y2.  When the user drags the mouse while using
 * the "Draw Curves" tool, a sequence of lines is produces.
 *    A stamp is one of a dozen images that are available in the
 * Tool menu.  When the user clicks the mouse while using a 
 * stamp tool, a copy of the stamp is placed at the point where
 * the user clicked.  The string for a stamp consists of
 * the word "stamp" followed by 3 numbers.  The numbers are
 * the number of the stamp and the x and y coordinates where
 * the stamp is placed.
 */
public class DrawPanel extends JPanel {
	
	private JLabel message;           // For displaying messages to the user.
	private Display display;          // The display area where the drawing is shown.
	private BufferedImage image;      // An off-screen copy of the drawing.
	private Graphics2D imageGraphics; // For drawing to the off-screen copy.
	private BufferedImage[] stamps;   // An array containing the 12 stamp images.
    private int tool = -1;            // Current tool; -1 means curve, >= 0 is a stamp number.
    private Color color;              // Current color for drawing lines.
    
	private DrawClient client;        // Represents the connection to the server.
	private int myConnectionID;       // Every client has an ID, assigned by the server; this is mine.

	
	/**
	 * This class defines the display area of the panel, where
	 * the drawing is shown.  All drawing is actually done to the
	 * off-screen copy, image.  The paintComponent() method in this
	 * class simply copies the image to the display.
	 */
	private class Display extends JPanel {
		protected void paintComponent(Graphics g) {
			synchronized( DrawPanel.this ) {
				g.drawImage(image,0,0,getWidth(),getHeight(),null);
			}
		}
	}
	
	
	/**
	 * This class defines the MouseListener that handles mouse clicking
	 * and dragging.  All the actual work of drawing is actually done by 
	 * calling the drawStamp() and drawLine() methods in the DrawPanel class.
	 */
	private class MouseHandler implements MouseListener, MouseMotionListener {
		private int prevX, prevY;
		private boolean dragging;
		public void mousePressed(MouseEvent evt) {
			if (dragging)
				return;
			if (tool >= 0) {
				drawStamp(tool, evt.getX(), evt.getY());
				return;
			}
			dragging = true;
			prevX = evt.getX();
			prevY = evt.getY();
		}
		public void mouseDragged(MouseEvent evt) {
			if (!dragging)
				return;
			int thisX = evt.getX();
			int thisY = evt.getY();
			drawLine(color, thisX, thisY, prevX, prevY);
			prevX = thisX;
			prevY = thisY;
		}
		public void mouseReleased(MouseEvent evt) {
			dragging = false;
		}
		public void mouseMoved(MouseEvent evt) { }
		public void mouseClicked(MouseEvent evt) { }
		public void mouseEntered(MouseEvent evt) { }
		public void mouseExited(MouseEvent evt) { }
	}
	

	/**
	 * This class defines the client object that represents the connection
	 * to the server.  It has a method that responds to messages received
	 * from the server by calling the netMessageReceived() method in the
	 * DrawPanel class.  The DrawPanel class has a variable named client
	 * of type DrawClient.  This variable has a method client.send(msg)
	 * that can be called to send msg to the server, which then forwards
	 * it to all the clients that are connected to the server.
	 */
	private class DrawClient extends Client {
		public DrawClient(String hubHostName, int hubPort) throws IOException {
			super(hubHostName, hubPort);
			myConnectionID = getID();
		}
		protected void messageReceived(Object message) {
			if (message instanceof ForwardedMessage) {
				ForwardedMessage msg = (ForwardedMessage)message;
				if (msg.message instanceof String)
					netMessageReceived((String)msg.message, msg.senderID);
			}
		}
	}
	
	
	/**
	 * The DrawPanel constructor asks the user to enter the IP address
	 * or host name of the server.  It then opens a connection to the
	 * server, which will be represented by the instance variable named
	 * "client".  If an error occurs, or if the user cancels, the
	 * program ends.  Otherwise, the panel is created with a large
	 * display area and a message lable at the bottom.
	 */
	public DrawPanel() {
		// TODO:  Create a client!
		display = new Display();
		display.setPreferredSize(new Dimension(800,600));
		message = new JLabel("Connected to the server as client number " + myConnectionID);
		message.setBackground(Color.LIGHT_GRAY);
		message.setOpaque(true);
		message.setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		image = new BufferedImage(800, 600, BufferedImage.TYPE_INT_ARGB);
		imageGraphics = image.createGraphics();
		imageGraphics.setColor(Color.WHITE);
		imageGraphics.fillRect(0, 0, 800, 600);
		imageGraphics.setColor(Color.BLACK);
		color = Color.BLACK;
		imageGraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
		imageGraphics.setStroke( new BasicStroke(2) );
		stamps = new BufferedImage[12];
		for (int i = 0; i < 12; i++) {
			try {
				String resname = "netgame/draw/stamps/icon" + i + ".png";
				URL resloc = getClass().getClassLoader().getResource(resname);
				stamps[i] = ImageIO.read(resloc);
			}
			catch (Exception e) {
			}
		}
		setLayout(new BorderLayout(3,3));
		setBackground(Color.GRAY);
		setBorder(BorderFactory.createEmptyBorder(3,3,3,3));
		add(display,BorderLayout.CENTER);
		add(message,BorderLayout.SOUTH);
		MouseHandler hndl = new MouseHandler();
		addMouseListener(hndl);
		addMouseMotionListener(hndl);
	}
	
	
	/**
	 * Creates a menu bar containing just a Tool menu.
	 * @return the menu bar that was created
	 */
	public JMenuBar createMenuBar() {
		JMenuBar menuBar = new JMenuBar();
		ActionListener lstn = new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				String cmd = evt.getActionCommand();
				if (cmd.equals("Draw Curves"))
					tool = -1;
				else if (cmd.startsWith("Stamp Number "))
					tool = Integer.parseInt(cmd.substring(13));
				else if (cmd.equals("Curve Color...")) {
					Color c = JColorChooser.showDialog(DrawPanel.this, "Select Curve Color", color);
					if (c != null)
						color = c;
				}
			}
		};
		JMenu menu = new JMenu("Tool");
		JMenuItem curves = new JMenuItem("Draw Curves");
		curves.addActionListener(lstn);
		menu.add(curves);
		JMenuItem colorItem = new JMenuItem("Curve Color...");
		colorItem.addActionListener(lstn);
		menu.add(colorItem);
		menu.addSeparator();
		for (int i = 0; i < 12; i++) {
			JMenuItem item = new JMenuItem("Stamp Number " + i);
			item.setIcon(new ImageIcon(stamps[i]));
			item.addActionListener(lstn);
			menu.add(item);
		}
		menuBar.add(menu);
		return menuBar;
	}
	
	
	/**
	 * This method is called when the user places a stamp in the drawing.
	 * It draws the stamp to the image in this panel and sends a message
	 * to the server about the stamp.  NOTE: This method should NOT be
	 * called in response to a message from the server, since that would
	 * result in ANOTHER message being sent back to the server!
	 * @param stampNumber The index of the stamp in the array of available stamps
	 * @param x the x-coordinate where the stamp is to be placed
	 * @param y the x-coordinate where the stamp is to be placed
	 */
	synchronized private void drawStamp(int stampNumber, int x, int y) {
		BufferedImage img = stamps[stampNumber];
		imageGraphics.drawImage(img, x - img.getWidth()/2, y - img.getHeight()/2, null);
		display.repaint();
		// TODO:  send this drawing operation over the network
	}
	
	
	/**
	 * This method when the user drags the mouse from one point to another.
	 * A line is drawn between the two points in the specified color, and
	 * a message is sent to the server about the line.  NOTE: This method
	 * should NOT be called in response to a message from the server, since 
	 * that wouldresult in ANOTHER message being sent back to the server!
	 * @param lineColor the color of the line
	 * @param x1 x-coord of the first endpoint
	 * @param y1 y-coord of the first endpoint
	 * @param x2 x-coord of the second endpoint
	 * @param y2 x-coord of the second endpoint
	 */
	synchronized private void drawLine(Color lineColor, int x1, int y1, int x2, int y2) {
		Rectangle rect = new Rectangle(x1,y1);
		rect.add(x2,y2);
		rect.grow(2,2);
		imageGraphics.setColor(lineColor);
		imageGraphics.drawLine(x1, y1, x2, y2);
		display.repaint(rect);
		// TODO:  send this drawing operation over the network
	}
	
	
	/**
	 * This method is called when a string is received as a message from
	 * the server.  The message actually come from one of the clients that
	 * are connected to the server.  Note then when the server gets a 
	 * message from THIS client, it sends a copy of that message back
	 * to this client!  Messages that originated from this client should
	 * be ignored because the graphics item represented by the message
	 * has already been drawn by this client.
	 * @param stringReceived the message from the server, which should
	 *    describe either a stamp or a line.
	 * @param senderID the ID number of the client who sent the message
	 *    to the server.  If this is the same as my ID, the message
	 *    should be ignored.
	 */
	synchronized private void netMessageReceived(String stringReceived, int senderID) {
		// TODO:  react to a message received from the network
	}
	
}
